@@ -1,6 +1,5 @@
 {-# LANGUAGE OverloadedStrings #-}
 {-# OPTIONS_GHC -Wno-name-shadowing #-}
-{-# OPTIONS_GHC -Wno-orphans #-}
 
 {- |
 Module      : GHC.Eventlog.Live.Otelcol
@@ -13,32 +12,35 @@ module GHC.Eventlog.Live.Otelcol (
 ) where
 
 import Control.Applicative (asum)
-import Control.Exception (Exception (..), catch)
 import Control.Monad.IO.Class (MonadIO (..))
 import Data.ByteString (ByteString)
+import Data.Coerce (Coercible, coerce)
 import Data.DList (DList)
 import Data.DList qualified as D
 import Data.Default (Default (..))
 import Data.Either (partitionEithers)
-import Data.Foldable (for_, traverse_)
+import Data.Foldable (for_)
 import Data.Functor ((<&>))
 import Data.HashMap.Strict qualified as M
 import Data.Hashable (Hashable)
-import Data.Int (Int64)
 import Data.Machine (MachineT, Process, ProcessT, asParts, await, construct, mapping, repeatedly, stopped, yield, (~>))
 import Data.Machine.Fanout (fanout)
 import Data.Maybe (fromMaybe, mapMaybe)
 import Data.ProtoLens (Message (defMessage))
+import Data.Proxy (Proxy (..))
+import Data.Semigroup (Last (..), Sum (..))
 import Data.Text (Text)
 import Data.Text qualified as T
 import Data.Text.Encoding qualified as TE
-import Data.Vector qualified as V
 import Data.Version (showVersion)
 import Data.Word (Word32, Word64)
 import Data.Yaml qualified as Y
+import GHC.Debug.Stub.Compat (MyGhcDebugSocket (..), maybeMyGhcDebugSocketParser, withMyGhcDebug)
 import GHC.Eventlog.Live.Data.Attribute
-import GHC.Eventlog.Live.Data.Metric
-import GHC.Eventlog.Live.Logger (logDebug, logError)
+import GHC.Eventlog.Live.Data.Group (Group, GroupBy, GroupedBy)
+import GHC.Eventlog.Live.Data.Group qualified as DG
+import GHC.Eventlog.Live.Data.Metric (Metric (..))
+import GHC.Eventlog.Live.Logger (logDebug)
 import GHC.Eventlog.Live.Machine.Analysis.Capability (CapabilityUsageSpan)
 import GHC.Eventlog.Live.Machine.Analysis.Capability qualified as M
 import GHC.Eventlog.Live.Machine.Analysis.Heap (MemReturnData (..))
@@ -52,22 +54,22 @@ import GHC.Eventlog.Live.Machine.WithStartTime qualified as M
 import GHC.Eventlog.Live.Options
 import GHC.Eventlog.Live.Otelcol.Config (Config)
 import GHC.Eventlog.Live.Otelcol.Config qualified as C
+import GHC.Eventlog.Live.Otelcol.Config.Default.Raw (defaultConfigString)
+import GHC.Eventlog.Live.Otelcol.Exporter (exportResourceMetrics, exportResourceSpans)
+import GHC.Eventlog.Live.Otelcol.Stats (Stat (..), eventCountTick, processStats)
 import GHC.Eventlog.Live.Socket (runWithEventlogSource)
 import GHC.Eventlog.Live.Verbosity (Verbosity)
+import GHC.Eventlog.Socket qualified as Eventlog.Socket
 import GHC.RTS.Events (Event (..), HeapProfBreakdown (..), ThreadId)
 import GHC.Records (HasField)
-import Lens.Family2 ((&), (.~), (^.))
+import Lens.Family2 ((&), (.~))
 import Network.GRPC.Client qualified as G
-import Network.GRPC.Client.StreamType.IO qualified as G
 import Network.GRPC.Common qualified as G
-import Network.GRPC.Common.Protobuf (Protobuf)
-import Network.GRPC.Common.Protobuf qualified as G
 import Options.Applicative qualified as O
 import Options.Applicative.Compat qualified as OC
 import Options.Applicative.Extra qualified as OE
 import Paths_eventlog_live_otelcol qualified as EventlogLive
 import Proto.Opentelemetry.Proto.Collector.Metrics.V1.MetricsService qualified as OMS
-import Proto.Opentelemetry.Proto.Collector.Metrics.V1.MetricsService_Fields qualified as OMS
 import Proto.Opentelemetry.Proto.Collector.Trace.V1.TraceService qualified as OTS
 import Proto.Opentelemetry.Proto.Collector.Trace.V1.TraceService_Fields qualified as OTS
 import Proto.Opentelemetry.Proto.Common.V1.Common qualified as OC
@@ -78,7 +80,6 @@ import Proto.Opentelemetry.Proto.Trace.V1.Trace qualified as OT
 import Proto.Opentelemetry.Proto.Trace.V1.Trace_Fields qualified as OT
 import System.Random (StdGen, initStdGen)
 import System.Random.Compat (uniformByteString)
-import Text.Printf (printf)
 
 {- |
 The main function for @eventlog-live-otelcol@.
@@ -86,85 +87,75 @@ The main function for @eventlog-live-otelcol@.
 main :: IO ()
 main = do
   Options{..} <- O.execParser options
-  let OpenTelemetryCollectorOptions{..} = openTelemetryCollectorOptions
-  -- Define the error writer:
-  let errorWriter = repeatedly $ await >>= logError verbosity . T.pack
-  -- Read the configuration file:
-  config <- flip (maybe (pure def)) maybeConfigFile $ \configFile -> do
-    logDebug verbosity $ "Reading configuration file from " <> T.pack configFile
-    config <- C.readConfig configFile
-    logDebug verbosity $ "Configuration file:\n" <> (TE.decodeUtf8Lenient . Y.encode $ config)
-    pure config
-  let attrServiceName = ("service.name", maybe AttrNull (AttrText . (.serviceName)) maybeServiceName)
-  G.withConnection G.def openTelemetryCollectorServer $ \conn -> do
-    runWithEventlogSource
-      verbosity
-      eventlogSocket
-      eventlogSocketTimeout
-      eventlogSocketTimeoutExponent
-      batchInterval
-      Nothing
-      maybeEventlogLogFile
-      $ fanout
-        [ M.validateInput verbosity 10
-        , M.counterByTick verbosity "events"
-        , M.liftTick M.withStartTime
-            ~> fanout
-              [ processHeapEvents config verbosity maybeHeapProfBreakdown
-                  ~> mapping (fmap Left)
-              , processThreadEvents config verbosity
-              ]
-            ~> mapping (partitionEithers . D.toList)
-            ~> fanout
-              [ runIf (shouldExportMetrics config) $
-                  mapping fst
-                    ~> fanout
-                      [ M.counterBy verbosity "metrics" (sum . fmap countMetricDataPoints)
-                      , asScopeMetrics
-                          [ OM.scope .~ eventlogLiveScope
-                          ]
-                          ~> asResourceMetric []
-                          ~> asExportMetricServiceRequest
-                          ~> exportResourceMetrics conn
-                          ~> mapping (either displayException displayException)
-                          ~> errorWriter
-                      ]
-              , runIf (shouldExportSpans config) $
-                  mapping snd
-                    ~> fanout
-                      [ M.counterBy verbosity "spans" (fromIntegral . length)
-                      , asScopeSpans
-                          [ OT.scope .~ eventlogLiveScope
-                          ]
-                          ~> asResourceSpan
-                            [ OT.resource
-                                .~ messageWith
-                                  [ OM.attributes .~ mapMaybe toMaybeKeyValue [attrServiceName]
-                                  ]
-                            ]
-                          ~> asExportTraceServiceRequest
-                          ~> exportResourceSpans conn
-                          ~> mapping (either displayException displayException)
-                          ~> errorWriter
-                      ]
-              ]
-        ]
 
-countMetricDataPoints :: OM.Metric -> Word
-countMetricDataPoints metric =
-  fromIntegral $
-    case metric ^. OM.maybe'data' of
-      Nothing -> 0
-      Just (OM.Metric'Gauge gauge) ->
-        V.length (gauge ^. OM.vec'dataPoints)
-      Just (OM.Metric'Sum sum) ->
-        V.length (sum ^. OM.vec'dataPoints)
-      Just (OM.Metric'Histogram histogram) ->
-        V.length (histogram ^. OM.vec'dataPoints)
-      Just (OM.Metric'ExponentialHistogram exponentialHistogram) ->
-        V.length (exponentialHistogram ^. OM.vec'dataPoints)
-      Just (OM.Metric'Summary summary) ->
-        V.length (summary ^. OM.vec'dataPoints)
+  -- Instument THIS PROGRAM with eventlog-socket and/or ghc-debug.
+  let MyDebugOptions{..} = myDebugOptions
+  withMyEventlogSocket maybeMyEventlogSocket
+  withMyGhcDebug verbosity maybeMyGhcDebugSocket $ do
+    --
+    -- Read the configuration file.
+    config <- flip (maybe (pure def)) maybeConfigFile $ \configFile -> do
+      logDebug verbosity $ "Reading configuration file from " <> T.pack configFile
+      config <- C.readConfig configFile
+      logDebug verbosity $ "Configuration file:\n" <> (TE.decodeUtf8Lenient . Y.encode $ config)
+      pure config
+
+    -- Create the service name attribute.
+    let attrServiceName = ("service.name", maybe AttrNull (AttrText . (.serviceName)) maybeServiceName)
+
+    -- Open a connection to the OpenTelemetry Collector.
+    let OpenTelemetryCollectorOptions{..} = openTelemetryCollectorOptions
+    G.withConnection G.def openTelemetryCollectorServer $ \conn -> do
+      runWithEventlogSource
+        verbosity
+        eventlogSocket
+        eventlogSocketTimeout
+        eventlogSocketTimeoutExponent
+        batchInterval
+        Nothing
+        maybeEventlogLogFile
+        $ fanout
+          [ M.validateInput verbosity 10
+          , eventCountTick ~> mapping (D.singleton . EventCountStat)
+          , M.liftTick M.withStartTime
+              ~> fanout
+                [ processHeapEvents config verbosity maybeHeapProfBreakdown
+                    ~> mapping (fmap Left)
+                , processThreadEvents config verbosity
+                ]
+              ~> mapping (partitionEithers . D.toList)
+              ~> fanout
+                [ runIf (shouldExportMetrics config) $
+                    mapping fst
+                      ~> fanout
+                        [ asScopeMetrics
+                            [ OM.scope .~ eventlogLiveScope
+                            ]
+                            ~> asResourceMetric []
+                            ~> asExportMetricServiceRequest
+                            ~> exportResourceMetrics conn
+                            ~> mapping (D.singleton . ExportMetricsResultStat)
+                        ]
+                , runIf (shouldExportSpans config) $
+                    mapping snd
+                      ~> fanout
+                        [ asScopeSpans
+                            [ OT.scope .~ eventlogLiveScope
+                            ]
+                            ~> asResourceSpan
+                              [ OT.resource
+                                  .~ messageWith
+                                    [ OM.attributes .~ mapMaybe toMaybeKeyValue [attrServiceName]
+                                    ]
+                              ]
+                            ~> asExportTraceServiceRequest
+                            ~> exportResourceSpans conn
+                            ~> mapping (D.singleton . ExportTraceResultStat)
+                        ]
+                ]
+          ]
+          ~> asParts
+          ~> processStats verbosity stats batchInterval 10
 
 {- |
 Internal helper.
@@ -233,11 +224,9 @@ processThreadEvents config verbosity =
             )
             ~> fanout
               [ runIf (C.processorEnabled (.metrics) (.capabilityUsage) config) $
-                  M.liftTick
-                    ( M.processCapabilityUsageMetrics
-                        ~> asNumberDataPoint
-                    )
-                    ~> M.batchByTickList
+                  M.liftTick M.processCapabilityUsageMetrics
+                    ~> aggregate viaSum (C.processorAggregationStrategy (.metrics) (.capabilityUsage) config)
+                    ~> mapping (fmap toNumberDataPoint)
                     ~> asSum
                       [ OM.aggregationTemporality .~ OM.AGGREGATION_TEMPORALITY_DELTA
                       , OM.isMonotonic .~ True
@@ -329,14 +318,74 @@ processHeapEvents config verbosity maybeHeapProfBreakdown =
     , processHeapProfSample config verbosity maybeHeapProfBreakdown
     ]
 
+--------------------------------------------------------------------------------
+-- Metric Aggregation
+
+data Aggregators a b = Aggregators
+  { nothing :: Process (Tick a) b
+  , byBatch :: Process (Tick a) b
+  }
+
+{- |
+Internal helper.
+Aggregate items based on the provided aggregators and aggregation strategy.
+-}
+aggregate :: Aggregators a b -> Maybe C.AggregationStrategy -> Process (Tick a) b
+aggregate Aggregators{..} = \case
+  Nothing -> nothing
+  Just C.AggregationStrategyByBatch -> byBatch
+
+{- |
+Internal helper.
+Metric aggregators via the `Semigroup` instance for `Sum`.
+-}
+viaSum :: forall a. (Num a) => Aggregators (Metric a) [Metric a]
+viaSum =
+  Aggregators
+    { nothing = M.batchByTickList
+    , byBatch =
+        -- TODO: emit group sample counts as separate metric
+        byBatchVia (Proxy @(Metric (Sum a)))
+          ~> mapping (fmap (.representative))
+    }
+
+{- |
+Internal helper.
+Metric aggregators via the `Semigroup` instance for `Last`.
+-}
+viaLast :: forall a. (GroupBy a) => Aggregators a [a]
+viaLast =
+  Aggregators
+    { nothing = M.batchByTickList
+    , byBatch =
+        -- TODO: emit group sample counts as separate metric
+        byBatchVia (Proxy @(Last a))
+          ~> mapping (fmap (.representative))
+    }
+
+{- |
+Internal helper.
+This function aggregates items via a `Semigroup` instance and grouped by the `GroupBy` instance.
+-}
+byBatchVia ::
+  forall a b.
+  (Coercible a b, GroupBy b, Semigroup b) =>
+  Proxy b ->
+  Process (Tick a) [Group a]
+byBatchVia (Proxy :: Proxy b) =
+  mapping (fmap DG.singleton . coerce @(Tick a) @(Tick b))
+    ~> M.aggregateByTick @(GroupedBy b)
+    ~> mapping (coerce @[Group b] @[Group a] . DG.groups)
+
 --------------------------------------------------------------------------------
 -- HeapAllocated
 
 processHeapAllocated :: Config -> Process (Tick (WithStartTime Event)) (DList OM.Metric)
 processHeapAllocated config =
   runIf (C.processorEnabled (.metrics) (.heapAllocated) config) $
-    M.liftTick (M.processHeapAllocatedData ~> asNumberDataPoint)
-      ~> M.batchByTickList
+    M.liftTick M.processHeapAllocatedData
+      ~> aggregate viaSum (C.processorAggregationStrategy (.metrics) (.heapAllocated) config)
+      ~> mapping (fmap toNumberDataPoint)
       ~> asSum
         [ OM.aggregationTemporality .~ OM.AGGREGATION_TEMPORALITY_DELTA
         , OM.isMonotonic .~ True
@@ -350,8 +399,9 @@ processHeapAllocated config =
 processHeapSize :: Config -> Process (Tick (WithStartTime Event)) (DList OM.Metric)
 processHeapSize config =
   runIf (C.processorEnabled (.metrics) (.heapSize) config) $
-    M.liftTick (M.processHeapSizeData ~> asNumberDataPoint)
-      ~> M.batchByTickList
+    M.liftTick M.processHeapSizeData
+      ~> aggregate viaLast (C.processorAggregationStrategy (.metrics) (.heapSize) config)
+      ~> mapping (fmap toNumberDataPoint)
       ~> asGauge
       ~> asMetricWith config (.heapSize) [OM.unit .~ "By"]
       ~> mapping D.singleton
@@ -362,8 +412,9 @@ processHeapSize config =
 processBlocksSize :: Config -> Process (Tick (WithStartTime Event)) (DList OM.Metric)
 processBlocksSize config =
   runIf (C.processorEnabled (.metrics) (.blocksSize) config) $
-    M.liftTick (M.processBlocksSizeData ~> asNumberDataPoint)
-      ~> M.batchByTickList
+    M.liftTick M.processBlocksSizeData
+      ~> aggregate viaLast (C.processorAggregationStrategy (.metrics) (.blocksSize) config)
+      ~> mapping (fmap toNumberDataPoint)
       ~> asGauge
       ~> asMetricWith config (.blocksSize) [OM.unit .~ "By"]
       ~> mapping D.singleton
@@ -374,8 +425,9 @@ processBlocksSize config =
 processHeapLive :: Config -> Process (Tick (WithStartTime Event)) (DList OM.Metric)
 processHeapLive config =
   runIf (C.processorEnabled (.metrics) (.heapLive) config) $
-    M.liftTick (M.processHeapLiveData ~> asNumberDataPoint)
-      ~> M.batchByTickList
+    M.liftTick M.processHeapLiveData
+      ~> aggregate viaLast (C.processorAggregationStrategy (.metrics) (.heapLive) config)
+      ~> mapping (fmap toNumberDataPoint)
       ~> asGauge
       ~> asMetricWith config (.heapLive) [OM.unit .~ "By"]
       ~> mapping D.singleton
@@ -387,20 +439,25 @@ processMemReturn :: Config -> Process (Tick (WithStartTime Event)) (DList OM.Met
 processMemReturn config =
   runIf (shouldComputeMemReturn config) $
     M.liftTick M.processMemReturnData
-      ~> M.batchByTickList
       ~> fanout
         [ runIf (C.processorEnabled (.metrics) (.memCurrent) config) $
-            M.liftBatch (asMemCurrent ~> asNumberDataPoint)
+            M.liftTick asMemCurrent
+              ~> aggregate viaLast (C.processorAggregationStrategy (.metrics) (.memCurrent) config)
+              ~> mapping (fmap toNumberDataPoint)
               ~> asGauge
               ~> asMetricWith config (.memCurrent) [OM.unit .~ "{mblock}"]
               ~> mapping D.singleton
         , runIf (C.processorEnabled (.metrics) (.memNeeded) config) $
-            M.liftBatch (asMemNeeded ~> asNumberDataPoint)
+            M.liftTick asMemNeeded
+              ~> aggregate viaLast (C.processorAggregationStrategy (.metrics) (.memNeeded) config)
+              ~> mapping (fmap toNumberDataPoint)
               ~> asGauge
               ~> asMetricWith config (.memNeeded) [OM.unit .~ "{mblock}"]
               ~> mapping D.singleton
         , runIf (C.processorEnabled (.metrics) (.memReturned) config) $
-            M.liftBatch (asMemReturned ~> asNumberDataPoint)
+            M.liftTick asMemReturned
+              ~> aggregate viaLast (C.processorAggregationStrategy (.metrics) (.memReturned) config)
+              ~> mapping (fmap toNumberDataPoint)
               ~> asGauge
               ~> asMetricWith config (.memReturned) [OM.unit .~ "{mblock}"]
               ~> mapping D.singleton
@@ -428,6 +485,20 @@ asMemReturned = mapping (fmap (.returned))
 --------------------------------------------------------------------------------
 -- HeapProfSample
 
+heapProfSampleDataAggregators :: Aggregators M.HeapProfSampleData [Metric Word64]
+heapProfSampleDataAggregators = Aggregators{..}
+ where
+  nothing :: Process (Tick M.HeapProfSampleData) [Metric Word64]
+  nothing =
+    M.batchByTickList
+      ~> mapping (concatMap M.heapProfSamples)
+  byBatch :: Process (Tick M.HeapProfSampleData) [Metric Word64]
+  byBatch =
+    -- TODO: emit group sample counts as separate metric
+    mapping (fmap (DG.singleton . Last))
+      ~> M.aggregateByTick @(GroupedBy (Last M.HeapProfSampleData))
+      ~> mapping (concatMap (M.heapProfSamples . getLast) . DG.elems)
+
 processHeapProfSample ::
   (MonadIO m) =>
   Config ->
@@ -436,8 +507,9 @@ processHeapProfSample ::
   ProcessT m (Tick (WithStartTime Event)) (DList OM.Metric)
 processHeapProfSample config verbosity maybeHeapProfBreakdown =
   runIf (C.processorEnabled (.metrics) (.heapProfSample) config) $
-    M.liftTick (M.processHeapProfSampleData verbosity maybeHeapProfBreakdown ~> asNumberDataPoint)
-      ~> M.batchByTickList
+    M.liftTick (M.processHeapProfSampleData verbosity maybeHeapProfBreakdown)
+      ~> aggregate heapProfSampleDataAggregators (C.processorAggregationStrategy (.metrics) (.heapProfSample) config)
+      ~> mapping (fmap toNumberDataPoint)
       ~> asGauge
       ~> asMetricWith config (.heapProfSample) [OM.unit .~ "By"]
       ~> mapping D.singleton
@@ -528,9 +600,6 @@ toSum mod dataPoints
   | null dataPoints = Nothing
   | otherwise = Just . OM.Metric'Sum . messageWith $ (OM.dataPoints .~ dataPoints) : mod
 
-asNumberDataPoint :: (IsNumberDataPoint'Value v) => Process (Metric v) OM.NumberDataPoint
-asNumberDataPoint = mapping toNumberDataPoint
-
 --------------------------------------------------------------------------------
 -- Interpret data
 --------------------------------------------------------------------------------
@@ -670,7 +739,7 @@ toNumberDataPoint i =
     [ OM.maybe'value .~ Just (toNumberDataPoint'Value i.value)
     , OM.timeUnixNano .~ fromMaybe 0 i.maybeTimeUnixNano
     , OM.startTimeUnixNano .~ fromMaybe 0 i.maybeStartTimeUnixNano
-    , OM.attributes .~ mapMaybe toMaybeKeyValue i.attr
+    , OM.attributes .~ mapMaybe toMaybeKeyValue (toList i.attrs)
     ]
 
 toMaybeKeyValue :: Attr -> Maybe OC.KeyValue
@@ -704,94 +773,6 @@ toMaybeAnyValue = \case
 messageWith :: (Message msg) => [msg -> msg] -> msg
 messageWith = foldr ($) defMessage
 
---------------------------------------------------------------------------------
--- OpenTelemetry Collector exporters
---------------------------------------------------------------------------------
-
---------------------------------------------------------------------------------
--- OpenTelemetry Collector Trace Exporter
-
-data ExportTraceError
-  = ExportTraceError
-  { rejectedSpans :: Int64
-  , errorMessage :: Text
-  }
-  deriving (Show)
-
-instance Exception ExportTraceError where
-  displayException :: ExportTraceError -> String
-  displayException ExportTraceError{..} =
-    printf "Error: OpenTelemetry Collector rejected %d data points with message: %s" rejectedSpans errorMessage
-
-sendResourceSpans :: G.Connection -> OTS.ExportTraceServiceRequest -> IO (Maybe (Either G.GrpcError ExportTraceError))
-sendResourceSpans conn exportTraceServiceRequest =
-  fmap (fmap Right) doGrpc `catch` handleGrpcError
- where
-  doGrpc :: IO (Maybe ExportTraceError)
-  doGrpc = do
-    G.nonStreaming conn (G.rpc @(Protobuf OTS.TraceService "export")) (G.Proto exportTraceServiceRequest) >>= \case
-      G.Proto resp
-        | resp ^. OTS.partialSuccess . OTS.rejectedSpans == 0 -> pure Nothing
-        | otherwise -> pure $ Just ExportTraceError{..}
-       where
-        rejectedSpans = resp ^. OTS.partialSuccess . OTS.rejectedSpans
-        errorMessage = resp ^. OTS.partialSuccess . OTS.errorMessage
-  handleGrpcError :: G.GrpcError -> IO (Maybe (Either G.GrpcError ExportTraceError))
-  handleGrpcError = pure . pure . Left
-
-exportResourceSpans :: G.Connection -> ProcessT IO OTS.ExportTraceServiceRequest (Either G.GrpcError ExportTraceError)
-exportResourceSpans conn =
-  repeatedly $
-    await >>= \exportTraceServiceRequest ->
-      liftIO (sendResourceSpans conn exportTraceServiceRequest)
-        >>= traverse_ yield
-
-type instance G.RequestMetadata (Protobuf OTS.TraceService meth) = G.NoMetadata
-type instance G.ResponseInitialMetadata (Protobuf OTS.TraceService meth) = G.NoMetadata
-type instance G.ResponseTrailingMetadata (Protobuf OTS.TraceService meth) = G.NoMetadata
-
---------------------------------------------------------------------------------
--- OpenTelemetry Collector Metrics Exporter
-
-data ExportMetricsError
-  = ExportMetricsError
-  { rejectedDataPoints :: Int64
-  , errorMessage :: Text
-  }
-  deriving (Show)
-
-instance Exception ExportMetricsError where
-  displayException :: ExportMetricsError -> String
-  displayException ExportMetricsError{..} =
-    printf "Error: OpenTelemetry Collector rejected %d data points with message: %s" rejectedDataPoints errorMessage
-
-sendResourceMetrics :: G.Connection -> OMS.ExportMetricsServiceRequest -> IO (Maybe (Either G.GrpcError ExportMetricsError))
-sendResourceMetrics conn exportMetricsServiceRequest =
-  fmap (fmap Right) doGrpc `catch` handleGrpcError
- where
-  doGrpc :: IO (Maybe ExportMetricsError)
-  doGrpc = do
-    G.nonStreaming conn (G.rpc @(Protobuf OMS.MetricsService "export")) (G.Proto exportMetricsServiceRequest) >>= \case
-      G.Proto resp
-        | resp ^. OMS.partialSuccess . OMS.rejectedDataPoints == 0 -> pure Nothing
-        | otherwise -> pure $ Just ExportMetricsError{..}
-       where
-        rejectedDataPoints = resp ^. OMS.partialSuccess . OMS.rejectedDataPoints
-        errorMessage = resp ^. OMS.partialSuccess . OMS.errorMessage
-  handleGrpcError :: G.GrpcError -> IO (Maybe (Either G.GrpcError ExportMetricsError))
-  handleGrpcError = pure . pure . Left
-
-exportResourceMetrics :: G.Connection -> ProcessT IO OMS.ExportMetricsServiceRequest (Either G.GrpcError ExportMetricsError)
-exportResourceMetrics conn =
-  repeatedly $
-    await >>= \exportMetricsServiceRequest ->
-      liftIO (sendResourceMetrics conn exportMetricsServiceRequest)
-        >>= traverse_ yield
-
-type instance G.RequestMetadata (Protobuf OMS.MetricsService meth) = G.NoMetadata
-type instance G.ResponseInitialMetadata (Protobuf OMS.MetricsService meth) = G.NoMetadata
-type instance G.ResponseTrailingMetadata (Protobuf OMS.MetricsService meth) = G.NoMetadata
-
 --------------------------------------------------------------------------------
 -- Options
 --------------------------------------------------------------------------------
@@ -815,8 +796,10 @@ data Options = Options
   , maybeHeapProfBreakdown :: Maybe HeapProfBreakdown
   , maybeServiceName :: Maybe ServiceName
   , verbosity :: Verbosity
+  , stats :: Bool
   , maybeConfigFile :: Maybe FilePath
   , openTelemetryCollectorOptions :: OpenTelemetryCollectorOptions
+  , myDebugOptions :: MyDebugOptions
   }
 
 optionsParser :: O.Parser Options
@@ -830,8 +813,48 @@ optionsParser =
     <*> O.optional heapProfBreakdownParser
     <*> O.optional serviceNameParser
     <*> verbosityParser
+    <*> statsParser
     <*> O.optional configFileParser
     <*> openTelemetryCollectorOptionsParser
+    <*> myDebugOptionsParser
+
+--------------------------------------------------------------------------------
+-- Debug Options
+
+data MyDebugOptions = MyDebugOptions
+  { maybeMyEventlogSocket :: Maybe MyEventlogSocket
+  , maybeMyGhcDebugSocket :: Maybe MyGhcDebugSocket
+  }
+
+myDebugOptionsParser :: O.Parser MyDebugOptions
+myDebugOptionsParser =
+  OC.parserOptionGroup "Debug Options" $
+    MyDebugOptions
+      <$> O.optional myEventlogSocketParser
+      <*> maybeMyGhcDebugSocketParser
+
+--------------------------------------------------------------------------------
+-- My Eventlog Socket
+
+newtype MyEventlogSocket
+  = MyEventlogSocketUnix FilePath
+
+myEventlogSocketParser :: O.Parser MyEventlogSocket
+myEventlogSocketParser =
+  MyEventlogSocketUnix
+    <$> O.strOption
+      ( O.long "enable-my-eventlog-socket-unix"
+          <> O.metavar "SOCKET"
+          <> O.help "Enable the eventlog socket for this program on the given Unix socket."
+      )
+
+{- |
+Set @eventlog-socket@ as the eventlog writer.
+-}
+withMyEventlogSocket :: Maybe MyEventlogSocket -> IO ()
+withMyEventlogSocket maybeMyEventlogSocket =
+  for_ maybeMyEventlogSocket $ \(MyEventlogSocketUnix myEventlogSocket) ->
+    Eventlog.Socket.startWait myEventlogSocket
 
 --------------------------------------------------------------------------------
 -- Configuration
@@ -846,13 +869,10 @@ configFileParser =
 
 defaultsPrinter :: O.Parser (a -> a)
 defaultsPrinter =
-  O.infoOption defaults . mconcat $
+  O.infoOption defaultConfigString . mconcat $
     [ O.long "print-defaults"
     , O.help "Print default configuration options that can be used in config.yaml"
     ]
- where
-  defaults :: String
-  defaults = T.unpack (TE.decodeUtf8Lenient (Y.encode (def :: Config)))
 
 --------------------------------------------------------------------------------
 -- Service Name
